from queue import Empty
from threading import Thread

import paramiko
from os import path


class RemoteExecutor:
    def __init__(self, task_queue, hosts):
        self.task_queue = task_queue
        self.hosts = hosts
        self.workers = None
        self.initialize_workers()

    def initialize_workers(self):
        self.workers = [SshWorker(host, self.task_queue) for host in self.hosts]

    def start(self):
        for worker in self.workers:
            worker.start()

    def wait(self):
        for worker in self.workers:
            worker.join()


class SshWorker(Thread):
    def __init__(self, host, task_queue):
        super(SshWorker, self).__init__()
        self.task_queue = task_queue
        self.host = host

    def run(self):
        client = spawn_ssh_client(self.host)
        while True:
            try:
                internal_task = self.task_queue.get(block=False)
            except Empty:
                break

            execute_remote_task(client, internal_task)


def execute_remote_task(client, internal_task):
    task = internal_task.task
    internal_task.future.set_running_or_notify_cancel()

    try:
        stdin, stdout, stderr = client.exec_command(task.command)

        if task.input:
            stdin.write(task.input)
            stdin.flush()

        task.output = ''.join(stdout.readlines())
        task.error = ''.join(stderr.readlines())

        internal_task.future.set_result(task)
    except Exception as e:
        internal_task.future.set_exception(e)


def spawn_ssh_client(host):
    key = None

    if host.key_file_path:
        key = paramiko.RSAKey.from_private_key_file(path.expanduser(host.key_file_path))

    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    client.connect(hostname=host.hostname,
                   password=host.password,
                   username=host.username,
                   port=host.port,
                   pkey=key)

    return client



from queue import Empty
from subprocess import run, PIPE
from threading import Thread


class LocalExecutor:
    def __init__(self, task_queue, thread_count):
        self.task_queue = task_queue
        self.thread_count = thread_count
        self.workers = None
        self.initialize_workers()

    def initialize_workers(self):
        self.workers = [LocalWorker(self.task_queue) for _ in range(self.thread_count)]

    def start(self):
        for worker in self.workers:
            worker.start()

    def wait(self):
        for worker in self.workers:
            worker.join()


class LocalWorker(Thread):
    def __init__(self, task_queue):
        super(LocalWorker, self).__init__()
        self.task_queue = task_queue

    def run(self):
        while True:
            try:
                internal_task = self.task_queue.get(block=False)
                execute_task(internal_task)
            except Empty:
                break


def execute_task(internal_task):
    task = internal_task.task
    internal_task.future.set_running_or_notify_cancel()

    try:
        completed_process = run([task.command], input=task.input, stdout=PIPE, stderr=PIPE,
                                timeout=task.time_limit, shell=True)
        task.output = completed_process.stdout.decode('utf-8')
        task.error = completed_process.stderr.decode('utf-8')

        internal_task.future.set_result(task)
    except Exception as e:
        internal_task.future.set_exception(e)



from distutils.core import setup
import setuptools

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='distlre',
    version='0.0.4',
    description='A Python package to distribute commands on remote hosts via'
                'SSH and to execute them locally in parallel.',
    long_description=long_description,
    url='https://github.com/csbence/DistLRE',
    author='Bence Cserna',
    packages=setuptools.find_packages(".", include=("distlre*",)),
    install_requires=['paramiko']
)



from concurrent.futures import Future
from queue import Queue


class Task:
    def __init__(self, command, meta, time_limit, memory_limit):
        self.meta = meta
        self.command = command
        self.input = None
        self.output = None
        self.error = None
        self.time_limit = time_limit
        self.memory_limit = memory_limit


class InternalTask:
    def __init__(self, task):
        self.task = task
        self.future = Future()


class RemoteHost:
    def __init__(self, hostname, port=22, username=None, password=None, key_file_path=None):
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.key_file_path = key_file_path


class DistLRE:
    def __init__(self, local_threads=0, remote_hosts=None):
        if remote_hosts is None:
            remote_hosts = []
        self.local_threads = local_threads
        self.remote_hosts = remote_hosts
        self.task_queue = Queue()
        self.local_executor = None
        self.remote_executor = None

        if local_threads != 0:
            from distlre.localexecutor import LocalExecutor
            self.local_executor = LocalExecutor(self.task_queue, local_threads)

        if remote_hosts:
            from distlre.remoteexecutor import RemoteExecutor
            self.remote_executor = RemoteExecutor(self.task_queue, remote_hosts)

    def submit(self, task):
        internal_task = InternalTask(task)

        self.task_queue.put(internal_task)

        return internal_task.future

    def execute_tasks(self):
        if self.local_executor:
            self.local_executor.start()

        if self.remote_executor is not None:
            self.remote_executor.start()

    def wait(self):
        if self.local_executor is not None:
            self.local_executor.wait()

        if self.remote_executor is not None:
            self.remote_executor.wait()
